package adalovelace

import (
	"embed"
	"encoding/json"
	"fmt"
	"io/fs"
	"io/ioutil"
	"log"
	"os"
	"text/template"

	"github.com/Masterminds/sprig"
	"sigs.k8s.io/yaml"
)

type Provisioner struct {
	Config
	ConfigPath string
	DryRun     bool
}

//go:embed templates/*/*
var tplFS embed.FS

func New(configPath string, dryRun bool) Provisioner {

	provisioner := Provisioner{
		DryRun:     dryRun,
		ConfigPath: configPath,
	}

	if err := provisioner.readConfig(); err != nil {
		log.Fatal(err)
	}

	return provisioner
}

func (p Provisioner) ApplyStage(name string) error {
	if name == "provisionning" {
		return p.ProvisionStage()
	}

	if name == "setup" {
		return p.SetupStage()
	}

	if name == "bootstrap" {
		return p.BootstrapStage()
	}

	if name == "install" {
		return p.InstallStage()
	}

	return fmt.Errorf("stage %s not found", name)
}

func (p Provisioner) ListStages() []string {
	return []string{"provisionning", "setup", "bootstrap", "install"}
}

func (p Provisioner) CheckRequirements() error {
	versions := []Version{
		{
			Command:    "k0sctl",
			Args:       []string{"version"},
			Regex:      `version: (.*)`,
			Constraint: "0.17.x",
		},
		{
			Command:    "terraform",
			Args:       []string{"version"},
			Regex:      `Terraform (v.*)`,
			Constraint: "1.x.x",
		},
		{
			Command:    "ansible-playbook",
			Args:       []string{"--version"},
			Regex:      `\[core (.*)\]`,
			Constraint: "2.x.x",
		},
	}

	if err := CheckVersions(versions); err != nil {
		return err
	}

	return nil
}

func (p Provisioner) PrintSSHConfig() error {
	tpl, err := template.New("ssh.tpl").Funcs(sprig.TxtFuncMap()).ParseFS(tplFS, "templates/ssh/ssh.tpl")
	if err != nil {
		return err
	}
	return tpl.Execute(os.Stdout, p.Config)
}

func (p *Provisioner) readConfig() error {
	rawConfig, err := ioutil.ReadFile(p.ConfigPath)
	if err != nil {
		return err
	}

	rawJson, err := yaml.YAMLToJSON(rawConfig)
	if err != nil {
		return err
	}

	var config Config
	if err := json.Unmarshal(rawJson, &config); err != nil {
		return err
	}

	p.Config = config
	return nil
}

func (p *Provisioner) saveConfig() error {
	rawJson, err := json.Marshal(&p.Config)
	if err != nil {
		return err
	}

	rawConfig, err := yaml.JSONToYAML(rawJson)
	if err != nil {
		return err
	}

	return ioutil.WriteFile(p.ConfigPath, rawConfig, fs.ModeAppend)
}
