package adalovelace

import (
	"encoding/json"
	"fmt"
	"io/fs"
	"io/ioutil"
	"log"
	"os"
	"os/exec"
	"path/filepath"
	"strings"

	"sigs.k8s.io/yaml"
)

type AnsibleInventory map[string]AnsibleInventoryUnit

type AnsibleInventoryUnit struct {
	Hosts    map[string]AnsibleInventoryHost `json:"hosts,omitempty"`
	Children AnsibleInventory                `json:"children,omitempty"`
	Vars     map[string]interface{}          `json:"vars,omitempty"`
}

type AnsibleInventoryHost struct {
	Name  string   `json:"name"`
	Node  Node     `json:"node"`
	Disks []string `json:"node_disks"`
	PVS   []string `json:"node_pvs"`
}

func (p *Provisioner) ansibleStage(playbook string) error {
	log.Println("cleaning ansbile folder")
	if err := checkConfig(p.Config); err != nil {
		return err
	}
	if err := refreshAnsibleFolder(); err != nil {
		return err
	}
	if err := copyAnsiblePlaybook(); err != nil {
		return err
	}
	if err := generateAnsibleInventory(&p.Config); err != nil {
		return err
	}
	if p.DryRun {
		if err := execAnsiblePlaybook(playbook, "--check"); err != nil {
			return err
		}
	} else {
		if err := execAnsiblePlaybook(playbook); err != nil {
			return err
		}
	}
	return nil
}

func (p *Provisioner) SetupStage() error {
	return p.ansibleStage("playbook.yaml")
}

func (p *Provisioner) InstallStage() error {
	p.Config.SetDefaults()
	if err := p.saveConfig(); err != nil {
		return err
	}
	return p.ansibleStage("install.yaml")
}

func checkConfig(config Config) error {
	if config.API.PublicIP == "" {
		return fmt.Errorf("the api doesn't have a public ip (try running the provisionning stage)")
	}

	if config.Ingress.PublicIP == "" {
		return fmt.Errorf("the ingress doesn't have a public ip (try running the provisionning stage)")
	}

	for nodeName, node := range config.Nodes {
		if node.PublicIP == "" {
			return fmt.Errorf("the node %s doesn't have a public ip (try running the provisionning stage)", nodeName)
		}
	}

	return nil
}

func refreshAnsibleFolder() error {
	if err := os.RemoveAll("ansible"); err != nil && !os.IsNotExist(err) {
		return err
	}

	err := os.Mkdir("ansible", 0764)
	if err != nil && !os.IsExist(err) {
		return err
	}
	return nil
}

func generateAnsibleInventory(c *Config) error {
	inventory := make(AnsibleInventory)
	inventory["all"] = AnsibleInventoryUnit{
		Children: AnsibleInventory{
			"controller": AnsibleInventoryUnit{
				Hosts: make(map[string]AnsibleInventoryHost),
			},
			"worker": AnsibleInventoryUnit{
				Hosts: make(map[string]AnsibleInventoryHost),
			},
		},
		Vars: map[string]interface{}{
			"cluster": c,
			"secrets": map[string]interface{}{
				"grafanaPassword": c.GrafanaPassword(),
			},
		},
	}

	for nodeName, node := range c.Nodes {
		host := AnsibleInventoryHost{
			Name:  nodeName,
			Node:  node,
			Disks: []string{},
			PVS:   []string{},
		}
		for _, volume := range node.Volumes {
			if volume.VG {
				host.PVS = append(host.PVS, volume.Device)
			} else {
				host.Disks = append(host.Disks, volume.Device)
			}
		}
		inventory["all"].Children[node.Role].Hosts[node.PublicIP] = host
	}

	j, err := json.Marshal(&inventory)
	if err != nil {
		return err
	}

	i, err := yaml.JSONToYAML(j)
	if err != nil {
		return err
	}

	return ioutil.WriteFile("ansible/inventory.yaml", i, 0664)
}

func copyAnsiblePlaybook() error {
	const source = "templates/ansible"
	const destination = "ansible"
	return fs.WalkDir(tplFS, source, func(path string, d fs.DirEntry, err error) error {
		var relPath string = strings.Replace(path, source, "", 1)
		if relPath == "" {
			return nil
		}
		if d.IsDir() {
			return os.Mkdir(filepath.Join(destination, relPath), 0764)
		} else {
			var data, err1 = tplFS.ReadFile(filepath.Join(source, relPath))
			if err1 != nil {
				return err1
			}
			return ioutil.WriteFile(filepath.Join(destination, relPath), data, 0754)
		}
	})
}

func execAnsiblePlaybook(extraArgs ...string) error {
	args := []string{
		"-i", "inventory.yaml",
	}
	args = append(args, extraArgs...)
	cmd := exec.Command("ansible-playbook", args...)
	cmd.Stdout = os.Stdout
	cmd.Stderr = os.Stderr
	cmd.Dir = "ansible"
	return cmd.Run()
}
