package adalovelace

import (
	"bytes"
	"encoding/json"
	"fmt"
	"io/fs"
	"log"
	"os"
	"os/exec"
	"path/filepath"
	"strings"
	"text/template"

	"github.com/Masterminds/sprig"
	jsonpatch "github.com/evanphx/json-patch"
)

func (p *Provisioner) ProvisionStage() error {
	log.Println("cleaning terraform folder")
	if err := refreshTerraformFolder(); err != nil {
		return err
	}

	log.Println("generating terraform files")
	if err := generatingTerraformFiles(&p.Config); err != nil {
		return err
	}

	log.Println("executing terraform")
	if err := execTerraform("init"); err != nil {
		return err
	}
	if p.DryRun {
		if err := execTerraformWithToken("plan"); err != nil {
			return err
		}
	} else {
		if err := execTerraformWithToken("apply", "-auto-approve"); err != nil {
			return err
		}
	}

	c, err := loadTerraformOutput(p.Config)
	if err != nil {
		return err
	}
	p.Config = c

	if p.DryRun {
		return nil
	}

	return p.saveConfig()
}

func refreshTerraformFolder() error {
	err := os.Mkdir("terraform", 0764)
	if err != nil && !os.IsExist(err) {
		return err
	}

	files, err := filepath.Glob("terraform/*.tf")
	if err != nil {
		return err
	}

	for _, file := range files {
		if err := os.Remove(file); err != nil {
			return err
		}
	}

	return nil
}

func generatingTerraformFiles(c *Config) error {
	return fs.WalkDir(tplFS, "templates/terraform", func(path string, d fs.DirEntry, err error) error {
		if !d.IsDir() {
			tpl, err := template.New(d.Name()).Funcs(template.FuncMap{
				"printL": printL,
			}).Funcs(sprig.TxtFuncMap()).ParseFS(tplFS, path)
			if err != nil {
				return err
			}

			f, err := os.OpenFile("terraform/"+tpl.Name(), os.O_CREATE|os.O_WRONLY, 0664)
			if err != nil {
				return err
			}
			defer f.Close()

			return tpl.Execute(f, c)
		}
		return nil
	})

}

func printL(list []string) string {
	return fmt.Sprintf("[%s]", strings.Join(list, ","))
}

func execTerraformWithToken(arg ...string) error {
	arg = append(arg, "-var", "hcloud_token="+os.Getenv("HCLOUD_TOKEN"))
	return execTerraform(arg...)
}

func execTerraform(arg ...string) error {
	cmd := exec.Command("terraform", arg...)
	cmd.Stdout = os.Stdout
	cmd.Stderr = os.Stderr
	cmd.Dir = "terraform"
	return cmd.Run()
}

func loadTerraformOutput(c Config) (Config, error) {
	cmd := exec.Command("terraform", "output", "-json", "cluster")
	var out bytes.Buffer
	cmd.Stdout = &out
	cmd.Stderr = os.Stderr
	cmd.Dir = "terraform"
	if err := cmd.Run(); err != nil {
		return Config{}, err
	}

	before, err := json.Marshal(c)
	if err != nil {
		return Config{}, nil
	}

	after, err := jsonpatch.MergePatch(before, out.Bytes())
	if err != nil {
		return Config{}, err
	}

	var config Config
	if err := json.Unmarshal(after, &config); err != nil {
		return Config{}, nil
	}

	return config, nil
}
